import os
import sys
import typer

from endoscopie import controller
from endoscopie import utils
from endoscopie.logarithm import Logarithm
from pathlib import Path
from rich import print
from typing import Optional

__version__ = '0.1.2'

app = typer.Typer()

global yam

logger = Logarithm(__name__)


@app.command()
def run(target: Optional[Path] = typer.Option(..., "--config",
                                              help="Set YAML file in which the datas required for image verify.")):
    _init(__version__)
    message = "Initializing the Endoscopie..."
    logger.info(message)

    _check(target)
    _setenv(yam.get('openstack'))
    _provision(target)


@app.command()
def cleanup():
    destroy: bool = typer.confirm("Are you sure you want to delete server(s)?", abort=True)
    _init(__version__)
    _clean_up(destroy)


def _clean_up(destroy: bool):
    message = "Start cleaning up the resources you used for testing."
    logger.info(message)

    try:
        controller.cleanup(destroy)
        utils.unset_variables()
    except Exception as e:
        logger.error(f"Failed to delete server that were used for the test. - {e}")

    logger.info("Finished cleaning.")
    sys.exit(0)


def _init(version: str) -> None:
    os.system('clear')
    print(utils.description(version))

    return None


def _check(target: Optional[Path]):
    logger.debug("Loading configurations...")

    if not _validation(target):
        message = "Failed to load configuration file. Please check the YAML file."
        print(f"[red]{message}[/red]")
        logger.error(message)

        sys.exit(1)

    logger.debug("Configuration successfully loaded.")


def _validation(target: Optional[Path]) -> bool:
    is_valid = False

    if _exist_file(target):
        if _is_malformed(target):
            is_valid = True

    return is_valid


def _exist_file(target: Optional[Path]) -> bool:
    output, error = utils.check_file_exist(target)

    if error is not None:
        logger.error(error)
        return False

    return True


def _is_malformed(target: Optional[Path]) -> bool:
    global yam
    yam = utils.load_yaml_file(target)

    if not yam:
        logger.error("Invalid YAML file (empty)")
        return False

    if 'openstack' not in yam.keys():
        logger.error("Invalid YAML file (Openstack Credential information is missing)")
        print("No Openstack Credential Information.")
        return False

    return True


def _setenv(openstack: dict) -> None:
    utils.setenv(openstack)


def _provision(target: Optional[Path]) -> None:
    servers = controller.provision(target)
    controller.verify(servers)



def main() -> int:

    app()
    return 0


if __name__ == "__main__":
    sys.exit(main())


import logging
import os
import socket
import sys
import time
from enum import Enum

import typer
import yaml
from rich.table import Table

from endoscopie.ssh import EndoscopieSSHClient
from endoscopie.logarithm import Logarithm
from endoscopie import commando
from endoscopie import utils
from endoscopie import template_meta as meta
from endoscopie import terraform as tf_action

from paramiko.ssh_exception import ChannelException, BadHostKeyException, AuthenticationException, SSHException
from pathlib import Path
from rich import print as rprint
from rich import box
from rich.console import Console
from rich.panel import Panel

logger = Logarithm(__name__)
template_dir = Path(f"{os.path.dirname(os.path.abspath(__file__))}/templates")

# BACK1 = '\b'
LINE_UP = '\x1B[A'
# LINE_UP2 = '\x1B M'
LINE_CLEAR = '\x1B[2K'
GREEN = '\x1B[32m'

global results


class Result(Enum):
    PASSED = f"[bold green]PASSED[/bold green]"
    FAILED = f"[bold red]FAILED[/bold red]"


def provision(image_info: Path):
    logger.info("Initializing the backend...")
    with open(image_info) as file:
        images = yaml.load(file, Loader=yaml.FullLoader)

    tf_template_meta = meta.make(images)
    utils.render_terraform_template(template_dir, 'openstack-vm-tf.temp', tf_template_meta)

    logger.info("You may now begin working with Terraform.")
    out, err = tf_action.create(template_dir)

    if err.strip() != '':
        rprint(Panel(err.strip(), title="[bold red]Failed to create server(s).[/bold red] :boom:"))
        cleanup(True)

        raise typer.Abort()

    info = utils.get_server_info(template_dir, images)
    servers = info['servers']
    rprint(Panel('\n'.join([f" :computer: [bold green] \[{i['id']}][/bold green] {i['name']}" for i in servers]),
                    title="[green]Successfully created server(s).[/green]"))

    return info


def cleanup(destroy: bool):
    if destroy:
        tf_action.server_list(template_dir)
        logger.debug(template_dir)
        if tf_action.is_exist(os.path.abspath(template_dir)):
            out, err = tf_action.delete(template_dir)

            if err.strip() != '':
                rprint(
                    Panel.fit(err.strip(), title="[bold red]Failed to delete server(s).[/bold red] :boom:"))

                # @TODO 실패하면 인스턴스 삭제 제시도 추가
                raise typer.Abort()

            rprint(Panel(f" :ok: [bold green]Successfully deleted server(s).[/bold green]",
                        title="[green]Clean Up[/green]"))

        else:
            rprint(f"[red]The server(s) to clean up does not exist.[/red]")

    sys.exit(0)


def verify(target: dict):
    logger.info("Endoscopie will start verifying the server(s).")

    bastion_host = target['bastionHost']
    bastion_host_ip = bastion_host['ip']
    bastion_host_user = bastion_host['user']
    bastion_keypair_name = bastion_host['keypair']['name']
    bastion_keypair_path = bastion_host['keypair']['path']

    servers = target['servers']

    logging.getLogger("paramiko.transport").setLevel(logging.FATAL)

    # create ssh client to bastion host server
    bastion = EndoscopieSSHClient()
    bastion_pkey = bastion.get_key_object(bastion_keypair_path)
    bastion.connect(hostname=bastion_host_ip, username=bastion_host_user, keypair=bastion_pkey)

    bstin, bstout, bsterr = bastion.exec_command("hostname -I")
    bastion_private_ip = bstout.readline().strip().replace('\n', '')
    logger.debug(f"The bastion host's private ip: {bastion_private_ip}")

    bastion_transport = bastion.get_transport()
    source = (bastion_private_ip, 22)

    global results
    results = []

    for idx, server in enumerate(servers):
        private_ip = server['access_ip_v4']
        name = server['name']
        keypair_path = server['keypair']
        user = name.split('-')[1]

        destination = (private_ip, 22)

        instance = EndoscopieSSHClient()
        vm_pkey = instance.get_key_object(keypair_path)

        retry_cnt = 0
        retry_message = ""
        max = 60
        while retry_cnt < max:
            if retry_cnt > 0:
                retry_message = f"(Number of retries: {retry_cnt})"

            try:
                print(f"Connecting to {private_ip}:22 [{name}] ... {retry_message}")
                channel = bastion_transport.open_channel("direct-tcpip", destination, source, timeout=120)
                instance.connect(hostname=private_ip, username=user, keypair=vm_pkey, sock=channel)
                break
            except ChannelException as ce:
                logger.error(f"Connection failed. No route to host.")
                time.sleep(1)
                _upAndClear(1)
                retry_cnt = utils.retry(retry_cnt)
                continue
            except BadHostKeyException as bhke:
                logger.error(f"Connection failed. The server's host key could not be verified.")
                time.sleep(1)
                _upAndClear(1)
                retry_cnt = utils.retry(retry_cnt)
                continue
            except AuthenticationException as ae:
                logger.error(f"Connection failed. (Authentication failed)")
                time.sleep(1)
                _upAndClear(1)
                retry_cnt = utils.retry(retry_cnt)
                continue
            except SSHException as se:
                logger.error(f"Connection failed. There was any other error connecting or establishing an SSH session.")
                time.sleep(1)
                _upAndClear(1)
                retry_cnt = utils.retry(retry_cnt)
                continue
            except socket.error as ske:
                logger.error(f"Connection failed. Could not connect to {private_ip}:22.")
                time.sleep(1)
                _upAndClear(1)
                retry_cnt = utils.retry(retry_cnt)
                continue

        is_passed = f"[bold green]{Result.PASSED.value}[/bold green]"

        if retry_cnt >= max - 1:
            rprint(f">>>>> [{name} ({private_ip})] Connection failed. There was any other error connecting or establishing an SSH session.")
            rprint(f"[bold red]\[FAILED][/bold red] Abort assertions.")
            is_passed = Result.FAILED.value
            results.append([idx + 1, server['id'], name, private_ip, is_passed])
            continue

        print(f"{LINE_UP}Connecting to {private_ip}:22 [{name}] ... {retry_message} => {GREEN}Connected.")
        rprint(f">>>>>>>> Start assertion - \[{name} ({private_ip})]")
        stdin, stdout, stderr = instance.exec_command(commando.cmd['hostname_s'])
        actual_hostname = stdout.readline().replace('\n', '')
        expect_hostname = 'host-' + private_ip.replace('.', '-')

        if expect_hostname != actual_hostname:
            is_passed = Result.FAILED.value
            rprint(f"""[bold red]\[FAILED][/bold red] The hostname should be equal to the expected hostname.
    - Expect hostname : {expect_hostname}
    - Actual hostname : {actual_hostname}""")
        else:
            rprint(f"[bold green]\[PASSED][/bold green] The hostname is '{actual_hostname}'")

        stdin, stdout, stderr = instance.exec_command(commando.cmd['hostname_i'])
        actual_ip = stdout.readline().strip().replace('\n', '')

        if actual_ip != private_ip:
            is_passed = Result.FAILED.value
            rprint(f"""
            [bold red]\[FAILED][/bold red] The private ip(v4) should be equal to the expected private ip.
                   - Expect IP (v4) : {private_ip}
                   - Actual IP (v4) : {actual_ip} 
            """)
        else:
            rprint(f"[bold green]\[PASSED][/bold green] The private ip(v4) is {actual_ip}")

        stdin, stdout, stderr = instance.exec_command(commando.cmd['vCPU'])
        expect_cpu = int(server['assertThat']['resources']['vCPU'])
        actual_cpu = int(stdout.readline().replace('\n', ''))

        if actual_cpu != expect_cpu:
            is_passed = Result.FAILED.value
            rprint(f"""[bold red]\[FAILED][/bold red] Server's number of vCPU should be equal to the expect value.
    - Expect vCPU : {expect_cpu} 개
    - Actual vCPU : {actual_cpu} 개""")
        else:
            rprint(f"[bold green]\[PASSED][/bold green] Server's number of vCPU is {actual_cpu}")

        stdin, stdout, stderr = instance.exec_command(commando.cmd['memory'])
        expect_memory = int(server['assertThat']['resources']['memory'])
        actual_memory = int(stdout.readline().replace('\n', ''))

        if actual_memory != expect_memory:
            is_passed = Result.FAILED.value
            rprint(f"""[bold red]\[FAILED][/bold red] Server's maximum capacity of memory should be equal to the expect value.
    - Expect maximum capacity of memory : {expect_memory} GiB
    - Actual maximum capacity of memory : {actual_memory} GiB""")
        else:
            rprint(f"[bold green]\[PASSED][/bold green] Server's maximum capacity of memory is {actual_memory} GiB")

        stdin, stdout, stderr = instance.exec_command(commando.cmd['disk'])
        expect_disk_size = server['assertThat']['resources']['volume']
        actual_disk_size = int(stdout.readline().replace('\n', ''))

        if _gib(actual_disk_size) != expect_disk_size:
            is_passed = Result.FAILED.value
            rprint(f"""[bold red]\[FAILED][/bold red] Server's volume(/dev/vda) size should be equal to the expect value.
    - Expect volume size : {expect_disk_size} GiB
    - Actual volume size : {_gib(actual_disk_size)} GiB""")
        else:
            rprint(f"[bold green]\[PASSED][/bold green] Server's volume(/dev/vda) size is {_gib(actual_disk_size)} GiB")

        stdin, stdout, stderr = instance.exec_command(commando.cmd['outbound'])
        actual_egress = stdout.readline().replace('\n', '').strip()

        if actual_egress == '000':
            is_passed = Result.FAILED.value
            rprint(f"""
            [bold red]\[FAILED][/bold red] Server's egress status should be equal to the expect value.
                   - Expect egress : Connection to Google(https://google.com) was successful.
                   - Actual egress : Failed to connect to Google(https://google.com).
            """)
        else:
            rprint(f"[bold green]\[PASSED][/bold green] Connection to Google(https://google.com) was successful")

        processes = server['assertThat']['processes']
        for process in processes:
            stdin, stdout, stderr = instance.exec_command(commando.cmd['process'].format(process))
            c = int(stdout.readline().replace('\n', '').strip())
            if c > 0:
                rprint(f"[bold green]\[PASSED][/bold green] The process is running. - '{process}'")
            else:
                is_passed = Result.FAILED.value
                rprint(f"[bold red]\[FAILED][/bold red] The process is not running. - '{process}'")

        ports = server['assertThat']['ports']
        for port in ports:
            stdin, stdout, stderr = instance.exec_command(commando.cmd['port'].format(int(port)))
            c = int(stdout.readline().replace('\n', '').strip())
            if c > 0:
                rprint(f"[bold green]\[PASSED][/bold green] The port is opened. - '{port}'")
            else:
                is_passed = Result.FAILED.value
                rprint(f"[bold red]\[FAILED][/bold red] The port is not opened. - '{port}'")

        results.append([idx + 1, server['id'], name, private_ip, is_passed])
        rprint("")
    _report(results)


def _report(data: list):
    console = Console()
    table = Table(show_header=True, header_style="bold magenta", title="View Test Report")
    table.width = console.width
    table.box = box.SIMPLE_HEAD
    table.pad_edge = False
    table.add_column("No", style="dim", width=5, justify="right")
    table.add_column("UUID")
    table.add_column("Server name")
    table.add_column("Private IP")
    table.add_column("Result")

    for value in data:
        table.add_row(str(value[0]), value[1], value[2], value[3], str(value[4]))

    console.print(table)

# @TODO
# def ssh_jump():
        # print(f"Assertion is complete.\n")

        # TO-DO
        # assert int(exported_port) == server['assertThat']['resources']['ports'], ""
        # assert processes

    # paramiko.common.logging.basicConfig(level=paramiko.common.DEBUG)

    # proxy_command = paramiko.ProxyCommand(
    #     f"ssh -i ~/set-shared-keypair.pem -W {vm_ip}:22 {bastion_host_user}@{bastion_host_ip}"
    # )
    # proxy.settimeout(300)
    # print(proxy_command.cmd)

    # vm_keypair = server['keypair']

    # print(f"-----> Connect to the bastion host")
    # pdb.set_trace()
    # ssh.connect(vm_ip, username=vm_user, allow_agent=True, pkey=private_key, sock=proxy_command)
    # print(f"-----> Connected to the bastion host")

    # connect to  virtual machine through bastion host
    # stdin, stdout, stderr = ssh.exec_command(f"hostname")
    # print(f">>> 1 >>> {stdout.readline()}")


#
# stdin, stdout, stderr = ssh.exec_command(f"ssh -i {vm_keypair} {vm_user}@{vm_ip}")
# print(f">>> 1 >>> {stdout.readline()}")
# stdin.write(f"sudo -i\n")
# print(f">>> 2 >>> {stdout.readline()}")
# stdin.write(f"hostname")
# print(f">>> 3 >>> {stdout.readline()}")

def _upAndClear(cnt):
    for idx in range(cnt):
        print(LINE_UP, end=LINE_CLEAR)


def _gib(size: int) -> int:
    return utils.gib(size)


if __name__ == '__main__':
    print(template_dir)




import io
import os.path

import paramiko
from endoscopie.logarithm import Logarithm
from paramiko.pkey import PKey

logger = Logarithm(__name__)


class EndoscopieSSHClient:
    def __init__(self):
        self.hostname = None
        self.username = None
        self.keypair = None
        self.sock = None

        self.ssh_client = paramiko.SSHClient()


    def get_key_object(self, keypair) -> PKey:
        self.ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        with open(os.path.abspath(keypair)) as keyfile:
            key = keyfile.read()

        private_key = paramiko.RSAKey.from_private_key(io.StringIO(key))
        return private_key


    def connect(self, hostname, username, keypair, sock=None) -> None:
        """
         :param hostname: str
             The instance IP(v4)
        :param username: str
            OS username
            ex)
               Ubuntu -> ubuntu
               CentOS and CentOS Stream -> centos
               Rocky Linux -> rocky
               Almalinux -> almalinux
        :param keypair: paramiko.pkey.PKey
            Path to the private key object for ssh connection
         :param socket sock:
            an open socket or socket-like object (such as a `.Channel`) to use for communication to the target host

        :return: None
        """
        logger.debug(f"hostname: {hostname}")
        logger.debug(f"username: {username}")
        self.hostname = hostname
        self.username = username
        self.keypair = keypair
        self.sock = sock

        if (self.hostname is None) or (self.username is None) or (self.keypair is None):
            logger.error("These values are required. ['hostname', 'username', 'keypair']")
            return

        self._connect_direct(self.hostname, self.username, self.keypair, self.sock)


    def exec_command(self, command):
        logger.debug(f"exec command : {command}")
        return self.ssh_client.exec_command(command=command)


    def get_transport(self):
        logger.debug("The `.Transport` object for this SSH connection.")
        return self.ssh_client.get_transport()


    def _connect_direct(self, hostname, username, keypair, sock):
        logger.debug("Connect to an SSH server and authenticate to it.")
        self.ssh_client.connect(hostname=hostname, username=username, pkey=keypair, sock=sock)






from endoscopie import utils
from endoscopie.logarithm import Logarithm

logger = Logarithm(__name__)


def create(template_dir):
    cmd_create = f"""
    bash -c "
    cd {template_dir}
    terraform init
    terraform plan
    terraform apply -auto-approve
    "
    """
    logger.info("Terraform will be created resources..")
    out, err = utils.execute_command(cmd_create)

    if err:
        logger.error(err)

    logger.debug(out)

    return out, err


def delete(template_dir):
    cmd_delete = f"""
    bash -c "
    cd {template_dir}
    pwd
    terraform init
    terraform plan
    terraform apply -destroy -auto-approve
    "
    """

    server_list(template_dir)
    logger.info("Terraform will be deleted resources..")
    out, err = utils.execute_command(cmd_delete)

    if err:
        logger.error(err)

    logger.debug(out)

    return out, err


def server_list(template_dir):
    cmd = f"""
    bash -c "
    cd {template_dir}
    terraform state list
    "
    """

    out, err = utils.execute_command(cmd)
    vm_server_list = out.strip().split("\n")

    # logger.debug(f"Server list : {vm_server_list}")

    return vm_server_list


def is_exist(template_dir):

    logger.debug(f"Terraform template path: {template_dir}")

    instances = server_list(template_dir)
    vm_server_list = [x for x in instances if x != '']

    if len(vm_server_list) <= 0:
        return False

    return True



import datetime
import logging
import logging.handlers
from pathlib import Path
from rich.logging import RichHandler


class Logarithm(logging.Logger):
    def __init__(self, name, level=logging.DEBUG):
        super().__init__(name, level)
        self.propagate = False
        """
        left align : [%(levelname)-8s]
        right align : [%(levelname)8s]
        """
        self.formatter = logging.Formatter(f"%(asctime)s [%(levelname)8s] --- %(message)s (%(filename)s:%(lineno)s)")
        self.levels = {
            'DEBUG': logging.DEBUG,
            'INFO': logging.INFO,
            'WARNING': logging.WARNING,
            'ERROR': logging.ERROR,
            'CRITICAL': logging.CRITICAL,
        }
        self.stream_handler(level="INFO")
        # self.rich_handler(level="INFO")
        self.rotate_file_handler(filename=f"{str(Path.home())}/.endoscopie/endoscopie.log",
                                 when='D',
                                 interval=1,
                                 backupCount=10,
                                 level="DEBUG")


    def stream_handler(self, level):
        stream = logging.StreamHandler()
        stream.setLevel(self.levels[level])
        stream.setFormatter(logging.Formatter("%(message)s"))
        self.addHandler(stream)


    def rich_handler(self, level):
        rich_handler = RichHandler()
        rich_handler.setLevel(self.levels[level])
        rich_handler.setFormatter(logging.Formatter("%(message)s"))
        self.addHandler(rich_handler)


    def file_handler(self, filename, mode, level):
        file = logging.FileHandler(filename=filename, mode=mode, encoding='utf-8')
        file.setLevel(self.levels[level])
        file.setFormatter(self.formatter)
        self.addHandler(file)


    def rotate_file_handler(self, filename, when, interval, backupCount, level):
        rotate = logging.handlers.TimedRotatingFileHandler(filename=filename,
                                                           when=when,
                                                           interval=interval,
                                                           backupCount=backupCount,
                                                           encoding='utf-8',
                                                           atTime=datetime.time(0, 0, 0))
        rotate.suffix = '%Y-%m-%d'
        rotate.setLevel(self.levels[level])
        rotate.setFormatter(self.formatter)
        self.addHandler(rotate)



import os
import uuid
from pathlib import Path
from endoscopie import utils
from endoscopie.logarithm import Logarithm

logger = Logarithm(__name__)


def make(images):
    metadata = utils.get_metadata(Path(f"{os.path.dirname(os.path.abspath(__file__))}/metadata/endoscopie-meta.json"))

    metadata["openstack"]["region"] = os.getenv("OS_REGION_NAME")
    metadata["openstack"]["auth"]["type"] = os.getenv("OS_AUTH_TYPE")
    metadata["openstack"]["auth"]["url"] = os.getenv("OS_AUTH_URL")
    metadata["openstack"]["credential"]["id"] = os.getenv("OS_APPLICATION_CREDENTIAL_ID")
    metadata["openstack"]["credential"]["secret"] = os.getenv("OS_APPLICATION_CREDENTIAL_SECRET")

    # with open(yaml_file) as file:
    #     images = yaml.load(file, Loader=yaml.FullLoader)

    servers = []
    for i in range(len(images["images"])):
        server = {
            "name": f'{images["images"][i]["osType"]}',
            "keypair": {
                "name": images["instance"]["keypair"]["name"]
            },
            "flavors": images["images"][i]["flavors"],
            "vpc": {
                "id": images["instance"]["vpc"]["id"],
                "subnet": images["instance"]["vpc"]["subnet"],
                "security_groups": images["instance"]["vpc"]["securityGroups"]
            },
            "block_device": {
                "uuid": images["images"][i]["id"],
                "source_type": "image",
                "volume_size": images["images"][i]["volumeSize"]
            },
            "image_id": images["images"][i]["id"],
            "image_name": images["images"][i]["name"],
            "os_type": images['images'][i]['osType'],
            "os_version": images['images'][i]['osVersion'],
            "os_user": images['images'][i]['osUser'],
            "user_script": {
                "path": images['images'][i]["userScript"]['path']
            # },
            # "assertThat": {
            #     "processes": images['images'][i]['assertThat']['processes'],
            #     "ports": images['images'][i]['assertThat']['ports']
                # "logs": images['images'][i]['assertThat']['logs'],
                # "files": images['images'][i]['assertThat']['files'],
                # "endpoint": {
                #     "url": images['images'][i]['assertThat']['endpoint']['url'],
                #     "id": images['images'][i]['assertThat']['endpoint']['id'],
                #     "password": images['images'][i]['assertThat']['endpoint']['password']
                # }
            }
        }

        servers.append(server)

    metadata["servers"] = servers

    logger.debug(metadata)

    return metadata



import json
import math
import os
import subprocess
import yaml

from endoscopie.logarithm import Logarithm
from jinja2 import Environment, FileSystemLoader
from pathlib import Path
from time import sleep
from typing import Optional

logger = Logarithm(__name__)


class CommandException(Exception):
    def __init__(self, error_message, return_code, stdout, stderr):
        self.error_message = error_message
        self.return_code = return_code
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        return f"{self.error_message} [return_code: {self.return_code}]"


def execute_command(command):
    try:
        process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        out, err = process.communicate()
    except Exception as e:
        logger.error(e)

    if process.returncode >= 1:
        logger.error(err.decode())
        raise CommandException(f"Command Execution Failed!", process.returncode, out.decode(), err.decode())

    return out.decode(), err.decode()


def render_terraform_template(tf_path, tf_file, metadata):
    env = Environment(loader=FileSystemLoader(tf_path))
    temp_tf = env.get_template(tf_file)
    terraform_tf = temp_tf.render(metadata)

    with open(f"{tf_path}/openstack-vm.tf", "w") as file:
        file.write(terraform_tf)


def get_metadata(meta_file):
    with open(meta_file) as mf:
        metadata = json.load(mf)
    return metadata


def get_server_info(template_dir, yaml_data):
    raw_data = get_resources(template_dir)

    servers = []
    resources = raw_data.get('values').get('root_module').get('resources')

    for resource in resources:
        flavor_id = resource.get('values').get('flavor_id')
        idx = int(resource.get('name').split('-')[-1]) - 1
        assert_that = yaml_data['images'][idx]['assertThat']

        sequence = 0
        for i, flavor in enumerate(yaml_data['images'][idx]['flavors']):
            if flavor_id != flavor['id']:
                continue
            sequence = i

        server = {
            "id": resource.get('values').get("id"),
            "name": resource.get('name'),
            "access_ip_v4": resource.get('values').get('access_ip_v4'),
            "access_ip_v6": resource.get('values').get('access_ip_v6'),
            "floating_ip": resource.get('values').get('floating_ip'),
            "keypair": yaml_data["instance"]["keypair"]["path"],
            "assertThat": {
                "resources": assert_that["resources"][sequence],
                "processes": assert_that["processes"],
                "ports": assert_that["ports"]
            }
        }

        servers.append(server)

    o = {
        "bastionHost": yaml_data["bastionHost"],
        "servers": servers
    }
    logger.debug(o)

    return o


def get_resources(template_dir) -> str:
    cmd = f"""
    bash -c "
    cd {template_dir}
    terraform show -json terraform.tfstate | jq
    " 
    """

    out, err = execute_command(cmd)

    return json.loads(out.strip())


def check_file_exist(file: Optional[Path]) -> tuple:
    error = None
    output = None

    if not file.exists():
        error = f":boom: [bold red]No such file or directory.[/bold red]"

        return output, error
    else:
        if file.is_file():
            if file.suffix not in ['.yaml', '.yml']:
                error = f":boom: [bold red] {file.absolute()}[/bold red] is not a [bold]YAML[/bold] file."
            # else:
            #     output = f":white_check_mark: [bold green] {file.absolute()}[/bold green]"
        elif file.is_dir():
            error = f":boom: [bold red]{file}[/bold red] is a directory."
        elif file.is_symlink():
            error = f":boom: [bold red]{file}[/bold red] is a symbolic link."
        elif file.is_mount():
            error = f":boom: [bold red]{file}[/bold red] is a POSIX mount point."
        elif file.is_block_device():
            error = f":boom: [bold red]{file}[/bold red] is a block device."
        elif file.is_char_device():
            error = f":boom: [bold red]{file}[/bold red] is a character device."
        elif file.is_socket():
            error = f":boom: [bold red]{file}[/bold red] is a socket."
        else:
            error = f":boom: [bold red]{file}[/bold red] is "

    return output, error


def unset_variables() -> None:
    cmd = f"""
    bash -c "
    unset OS_REGION_NAME
    unset OS_AUTH_URL
    unset OS_AUTH_TYPE
    unset OS_APPLICATION_CREDENTIAL_ID
    unset OS_APPLICATION_CREDENTIAL_SECRET
    "
    """

    out, err = execute_command(cmd)
    if err:
        logger.error(err.strip())

    logger.debug(out.strip())

    return None


def load_yaml_file(target: Optional[Path]) -> dict:
    logger.debug("Loading YAML file...")

    yml = {}
    try:
        with open(target, 'r') as file:
            yml = yaml.load(file, Loader=yaml.FullLoader)
    except IOError as ioe:
        logger.error(f"Loading the YAML file failed. - {ioe}")

    return yml


def _check_openstack_credential(openstack: dict) -> bool:
    malformed = False

    if 'openstack' in openstack.keys():
        malformed = True

    return malformed


def setenv(openstack: dict) -> None:
    logger.info("Setting environment variables...")

    os.environ['OS_REGION_NAME'] = openstack.get('OS_REGION_NAME')
    os.environ['OS_AUTH_URL'] = openstack.get('OS_AUTH_URL')
    os.environ['OS_AUTH_TYPE'] = openstack.get('OS_AUTH_TYPE')
    os.environ['OS_APPLICATION_CREDENTIAL_ID'] = openstack.get('OS_APPLICATION_CREDENTIAL_ID')
    os.environ['OS_APPLICATION_CREDENTIAL_SECRET'] = openstack.get('OS_APPLICATION_CREDENTIAL_SECRET')

    logger.debug(f"OS_REGION_NAME={os.getenv('OS_REGION_NAME')}")
    logger.debug(f"OS_AUTH_URL={os.getenv('OS_AUTH_URL')}")
    logger.debug(f"OS_AUTH_TYPE={os.getenv('OS_AUTH_TYPE')}")
    logger.debug(f"OS_APPLICATION_CREDENTIAL_ID={os.getenv('OS_APPLICATION_CREDENTIAL_ID')}")
    logger.debug(f"OS_APPLICATION_CREDENTIAL_SECRET={os.getenv('OS_APPLICATION_CREDENTIAL_SECRET')}")


def retry(cnt: int) -> int:
    sleep(3)
    var = cnt + 1
    return var


def gib(size: int) -> int:
    return int(size / math.pow(1024, 3))


def description(version):
    logo = f"""


$$$$$$$$\                 $$\                                                   $$\           
$$  _____|                $$ |                                                  \__|          
$$ |      $$$$$$$\   $$$$$$$ | $$$$$$\   $$$$$$$\  $$$$$$$\  $$$$$$\   $$$$$$\  $$\  $$$$$$\  
$$$$$\    $$  __$$\ $$  __$$ |$$  __$$\ $$  _____|$$  _____|$$  __$$\ $$  __$$\ $$ |$$  __$$\ 
$$  __|   $$ |  $$ |$$ /  $$ |$$ /  $$ |\$$$$$$\  $$ /      $$ /  $$ |$$ /  $$ |$$ |$$$$$$$$ |
$$ |      $$ |  $$ |$$ |  $$ |$$ |  $$ | \____$$\ $$ |      $$ |  $$ |$$ |  $$ |$$ |$$   ____|
$$$$$$$$\ $$ |  $$ |\$$$$$$$ |\$$$$$$  |$$$$$$$  |\$$$$$$$\ \$$$$$$  |$$$$$$$  |$$ |\$$$$$$$\ 
\________|\__|  \__| \_______| \______/ \_______/  \_______| \______/ $$  ____/ \__| \_______|
                                                                      $$ |                    
                                                                      $$ |                    
                                                                      \__|                    

                                                                            Version {version}
"""
    return logo


def mklogdir():
    app_path = f"{str(Path.home())}/.endoscopie"
    os.makedirs(app_path, exist_ok=True)




cmd = {
    "hostname_s": "hostname -s",
    "hostname_i": "hostname -I",
    "vCPU": "grep -c ^processor /proc/cpuinfo",
    "memory": "sudo dmidecode -t memory | grep 'Maximum Capacity' | cut -d':' -f2 | awk '{print $1}'",
    "disk": "sudo fdisk -l | grep 'Disk /dev/vda' | awk '{print $5}'",
    "outbound": "curl --connect-timeout 5 -s -o /dev/null -I -w '%{http_code}\n' https://www.google.com/",
    "process": "sudo pgrep -c {}",
    "port": "ss -lt src :{} | grep LISTEN | wc -l"
}